# select_toys.rb
#-----------------------------------------------------------------------------
# Version: 1.12.1 [DEBUG] DO NOT REDISTRIBUTE!
# Compatible: SketchUp 7 (PC)
#             SketchUp 6 (PC)
#             (other versions untested)
#-----------------------------------------------------------------------------
#
# Explaination of all features:
# http://www.sketchucation.com/forums/scf/viewtopic.php?f=180&t=14975
#
#-----------------------------------------------------------------------------
#
# Changelog
# 1.12.0 - 09.07.2009
#       * Added function to select all active entities on selected layers
#
# 1.11.0 - 09.07.2009
#       * Added Layer filters
#       * Added function to select all entities on selected layers
#
# 1.10.1 - 08.07.2009
#       * Forgot to add the menu items for 3D Polylines
#       * Fixed a bug in the drawingelement filter methods
#
# 1.10.0 - 08.07.2009
#       * Added code for Edge Loop (WIP) Not exposed UI.
#       * Added filters for 3D Polylines
#       * Modified circle filters to work around a bug where circles report an end angle of 720
#       * Modified circle filters to account for arches with start-angle other than 0 that makes up 360 degrees
#       * Corrected a flaw in the .typename optimisations.
#
# 1.9.0 - 11.06.2009
#       * Added filters that differs between ArcCurves and Curves
#       * Fixed a bug in selection Circles where Curves would cause an error
#
# 1.8.0 - 10.06.2009
#       * Optimized code that's using .typename
#       * Added Select Face Same Direction
#       * Added workaround for SU bug where group definition references 
#         becomes corrupted
#
# 1.7.1 - 20.05.2009
#       * Added selection of faces of the same area
#       * Added selection of opposite face
#
# 1.7.0 - 07.05.2009
#       * Added filters for Images
#
# 1.6.1 - 18.03.2009
#       * Added filters for Hidden, Soft and Smooth
#
# 1.6.0 - 16.03.2009
#       * Reorganised the selection filter methods
#       * Added support for DimensionRadial and DimensionLinear selection filter
#       * Added Select Parallel and Select Perpendicular faces
#
# 1.5.1 - 13.03.2009
#       * Fixed an issue where the DC menu would not appear
#
# 1.5.0 - 12.03.2009
#       * Added ability to select DC components by similar family
#
# 1.4.0 - 17.02.2009
#       * Filter/Deselect entities with default materials
#
# 1.3.2 - 23.01.2009
#       * Prevents Select Active * from selecting items on hidden layers
#
# 1.3.1 - 14.01.2009
#       * Eliminating Bugsplats due to recursing
#       * [WIP] Select Connected Faces by Angle
#       * Addressed issue of context menu graying out
#       * Select Active from same Layer
#       * Select All from same Layer
#
# 1.2.1 - 07.01.2009
#       * Selection manipulation (filters)
#       * Select Connected Coplanar Edges
#       * Select Connected by Material
#       * Select Connected by Layer
#
# 1.1.2 - 16.12.2008
#       * Select Loop from edge
#
# 1.0.0 - 13.12.2008
#       * Select Active Instances
#       * Select Instances
#       * Select Active Group Copies
#       * Select Group Copies
#       * Convert Group Copies to Instances
#
#-----------------------------------------------------------------------------
#
# Known Issues:
#  * Ghost selections appear when selecting all groups/instances and some of
#    them are nested inside other groups or components.
#    (Not fixable via ruby. Must be fixed in Sketchup.)
#
#  * Select Active * selects hidden geometry
#
#-----------------------------------------------------------------------------
#
# Todo:
#  * Improve Connected Face by Angle selection
#  * Add Organizer support
#  * Select active should select hidden only if hidden geometry is visible.
#  * Merge Group and Component select methods into the same methods.
#  * Make the Group, Component and DC select methods accept multiple selection.
#
#-----------------------------------------------------------------------------
#
# Thomas Thomassen
# thomas[at]thomthom[dot]net
#
#-----------------------------------------------------------------------------

require 'sketchup.rb'

#-----------------------------------------------------------------------------
# Add some menu items to access this
unless file_loaded?('select_toys.rb')
	# --- Commands --- #
	# Component
	cmd_select_active_instances 		= UI::Command.new('Select Active')					{ Select_Toys.select_active_instances() }
	cmd_select_instances				= UI::Command.new('Select All')						{ Select_Toys.select_instances() }
	cmd_select_active_instances_layer	= UI::Command.new('Select Active from same Layer')	{ Select_Toys.select_active_instances_from_same_layer() }
	cmd_select_instances_layer			= UI::Command.new('Select All from same Layer')		{ Select_Toys.select_instances_from_same_layer() }
	# Dynamic Component
	cmd_select_dc_active_instances 		= UI::Command.new('Select Active')					{ Select_Toys.select_dc_active_instances() }
	cmd_select_dc_instances				= UI::Command.new('Select All')						{ Select_Toys.select_dc_instances() }
	cmd_select_dc_active_instances_layer= UI::Command.new('Select Active from same Layer')	{ Select_Toys.select_dc_active_instances_from_same_layer() }
	cmd_select_dc_instances_layer		= UI::Command.new('Select All from same Layer')		{ Select_Toys.select_dc_instances_from_same_layer() }
	#Group
	cmd_select_active					= UI::Command.new('Select Active')					{ Select_Toys.select_active_group_copies() }
	cmd_select_all						= UI::Command.new('Select All')						{ Select_Toys.select_group_copies() }
	cmd_select_active_groups_layer		= UI::Command.new('Select Active from same Layer')	{ Select_Toys.select_active_group_copies_from_same_layer() }
	cmd_select_group_copies_layer		= UI::Command.new('Select All from same Layer')		{ Select_Toys.select_group_copies_from_same_layer() }
	cmd_convert_into_components			= UI::Command.new('Convert into Components')		{ Select_Toys.convert_group_copies_to_components() }
	# Edge
	cmd_loop_from_edge					= UI::Command.new('Quad-face Loop')					{ Select_Toys.select_loop_from_edge() }
	cmd_edge_loop						= UI::Command.new('Edge Loop')						{ Select_Toys.select_edge_loop() }
	# Face
	cmd_connected_perpendicular_faces	= UI::Command.new('Connected Perpendicular Faces')	{ Select_Toys.select_connected_perpendicular_faces() }
	cmd_connected_parallel_faces		= UI::Command.new('Connected Parallel Faces')		{ Select_Toys.select_connected_parallel_faces() }
	cmd_connected_planar_faces			= UI::Command.new('Connected Coplanar Faces')		{ Select_Toys.select_connected_planar_faces() }
	cmd_connected_faces_by_angle		= UI::Command.new('Connected Faces by Angle')		{ Select_Toys.tool(Select_Toys::Select_Connected_Faces_By_Angle_Tool.new) }
	cmd_connected_faces_by_area			= UI::Command.new('Connected Faces by Area')		{ Select_Toys.select_connected_same_area() }
	cmd_perpendicular_faces				= UI::Command.new('Perpendicular Faces')			{ Select_Toys.select_perpendicular_faces() }
	cmd_same_direction_faces			= UI::Command.new('Faces in Same Direction')		{ Select_Toys.select_same_direction_faces() }
	cmd_parallel_faces					= UI::Command.new('Parallel Faces')					{ Select_Toys.select_parallel_faces() }
	cmd_opposite_faces					= UI::Command.new('Opposite Faces')					{ Select_Toys.select_opposite_faces() }
	cmd_faces_by_area					= UI::Command.new('Faces by Area')					{ Select_Toys.select_same_area() }
	# Edge / Face
	cmd_connected_by_material			= UI::Command.new('Connected By Material')			{ Select_Toys.select_connected_by_material(false) }
	cmd_connected_by_back_material		= UI::Command.new('Connected By Back Material')		{ Select_Toys.select_connected_by_material(true) }
	cmd_connected_by_layer				= UI::Command.new('Connected By Layer')				{ Select_Toys.select_connected_by_layer() }
	# Entities
	cmd_select_active_by_selected_layers= UI::Command.new('Active on Selected Layers')		{ Select_Toys.select_active_by_selected_layers() }
	cmd_select_by_selected_layers		= UI::Command.new('All on Selected Layers')			{ Select_Toys.select_by_selected_layers() }
	# Filter Selection
	cmd_filter_edges					= UI::Command.new('Edges')							{ Select_Toys.filter_selection(Sketchup::Edge) }
	cmd_filter_faces					= UI::Command.new('Faces')							{ Select_Toys.filter_selection(Sketchup::Face) }
	cmd_filter_groups					= UI::Command.new('Groups')							{ Select_Toys.filter_selection(Sketchup::Group) }
	cmd_filter_components				= UI::Command.new('Components')						{ Select_Toys.filter_selection(Sketchup::ComponentInstance) }
	cmd_filter_guides					= UI::Command.new('Guides')							{ Select_Toys.filter_selection(Sketchup::ConstructionLine) }
	cmd_filter_points					= UI::Command.new('Construction Points')			{ Select_Toys.filter_selection(Sketchup::ConstructionPoint) }
	cmd_filter_text						= UI::Command.new('Text')							{ Select_Toys.filter_selection(Sketchup::Text) }
	cmd_filter_images					= UI::Command.new('Images')							{ Select_Toys.filter_selection(Sketchup::Image) }
	cmd_filter_curves					= UI::Command.new('Curves')							{ Select_Toys.filter_selection_curves() }
	cmd_filter_arcs						= UI::Command.new('Arcs')							{ Select_Toys.filter_selection_arc() }
	cmd_filter_circles					= UI::Command.new('Circles')						{ Select_Toys.filter_selection_circle() }
	cmd_filter_3dpoly					= UI::Command.new('3D Polyline')					{ Select_Toys.filter_selection_drawingelement('Polyline3d') }
	cmd_filter_dim_lin					= UI::Command.new('Linear Dimension')				{ Select_Toys.filter_selection_drawingelement('DimensionLinear') }
	cmd_filter_dim_rad					= UI::Command.new('Radial Dimension')				{ Select_Toys.filter_selection_drawingelement('DimensionRadial') }
	cmd_filter_default_front_material	= UI::Command.new('Front Default Material')			{ Select_Toys.filter_selection_default_material(false) }
	cmd_filter_default_back_material	= UI::Command.new('Back Default Material')			{ Select_Toys.filter_selection_default_material(true) }
	cmd_filter_hidden					= UI::Command.new('Hidden')							{ Select_Toys.filter_selection2() { |e| e.hidden?} }
	cmd_filter_soft						= UI::Command.new('Soft')							{ Select_Toys.filter_selection2() { |e| e.kind_of?(Sketchup::Edge) && e.soft?} }
	cmd_filter_smooth					= UI::Command.new('Smooth')							{ Select_Toys.filter_selection2() { |e| e.kind_of?(Sketchup::Edge) && e.smooth?} }
	# Deselect Selection
	cmd_deselect_edges					= UI::Command.new('Edges')							{ Select_Toys.deselect(Sketchup::Edge) }
	cmd_deselect_faces					= UI::Command.new('Faces')							{ Select_Toys.deselect(Sketchup::Face) }
	cmd_deselect_groups					= UI::Command.new('Groups')							{ Select_Toys.deselect(Sketchup::Group) }
	cmd_deselect_components				= UI::Command.new('Components')						{ Select_Toys.deselect(Sketchup::ComponentInstance) }
	cmd_deselect_guides					= UI::Command.new('Guides')							{ Select_Toys.deselect(Sketchup::ConstructionLine) }
	cmd_deselect_points					= UI::Command.new('Construction Points')			{ Select_Toys.deselect(Sketchup::ConstructionPoint) }
	cmd_deselect_text					= UI::Command.new('Text')							{ Select_Toys.deselect(Sketchup::Text) }
	cmd_deselect_images					= UI::Command.new('Images')							{ Select_Toys.deselect(Sketchup::Image) }
	cmd_deselect_curves					= UI::Command.new('Curves')							{ Select_Toys.deselect_curves() }
	cmd_deselect_arcs					= UI::Command.new('Arcs')							{ Select_Toys.deselect_arc() }
	cmd_deselect_circles				= UI::Command.new('Circles')						{ Select_Toys.deselect_circle() }
	cmd_deselect_3dpoly					= UI::Command.new('3D Polyline')					{ Select_Toys.deselect_drawingelement('Polyline3d') }
	cmd_deselect_dim_lin				= UI::Command.new('Linear Dimension')				{ Select_Toys.deselect_drawingelement('DimensionLinear') }
	cmd_deselect_dim_rad				= UI::Command.new('Radial Dimension')				{ Select_Toys.deselect_drawingelement('DimensionRadial') }
	cmd_deselect_default_front_material	= UI::Command.new('Front Default Material')			{ Select_Toys.deselect_default_material(false) }
	cmd_deselect_default_back_material	= UI::Command.new('Back Default Material')			{ Select_Toys.deselect_default_material(true) }
	cmd_deselect_hidden					= UI::Command.new('Hidden')							{ Select_Toys.deselect2() { |e| e.hidden?} }
	cmd_deselect_soft					= UI::Command.new('Soft')							{ Select_Toys.deselect2() { |e| e.kind_of?(Sketchup::Edge) && e.soft?} }
	cmd_deselect_smooth					= UI::Command.new('Smooth')							{ Select_Toys.deselect2() { |e| e.kind_of?(Sketchup::Edge) && e.smooth?} }
		
	UI.add_context_menu_handler do |menu|
		# --- Menus --- #
		menu.add_separator
		# Component
		if Select_Toys.selected_component?
			menu_instances = menu.add_submenu('Instances')
			menu_instances.add_item(cmd_select_active_instances)
			menu_instances.add_item(cmd_select_instances)
			menu_instances.add_item(cmd_select_active_instances_layer)
			menu_instances.add_item(cmd_select_instances_layer)
		end
		# Dynamic Component
		if Select_Toys.selected_dc_component?
			menu_dc_instances = menu.add_submenu('DC Instances')
			menu_dc_instances.add_item(cmd_select_dc_active_instances)
			menu_dc_instances.add_item(cmd_select_dc_instances)
			menu_dc_instances.add_item(cmd_select_dc_active_instances_layer)
			menu_dc_instances.add_item(cmd_select_dc_instances_layer)
		end
		# Group
		if Select_Toys.selected_group_copies?
			menu_copies = menu.add_submenu('Group Copies')
			menu_copies.add_item(cmd_select_active)
			menu_copies.add_item(cmd_select_all)
			menu_copies.add_item(cmd_select_active_groups_layer)
			menu_copies.add_item(cmd_select_group_copies_layer)
			menu_copies.add_item(cmd_convert_into_components)
		end
		# Entities
		if Sketchup.active_model.selection.length > 0
			menu_select = menu.add_submenu('Select')
			menu_select.add_item(cmd_select_active_by_selected_layers)
			menu_select.add_item(cmd_select_by_selected_layers)
			# Edge
			if Select_Toys.selected_edge?
				menu_select.add_separator()
				#menu_select.add_item(cmd_edge_loop)
				menu_select.add_item(cmd_loop_from_edge)
				menu_select.add_item(cmd_connected_by_material)
				menu_select.add_item(cmd_connected_by_layer)
			end
			# Face
			if Select_Toys.selected_face?
				menu_select.add_separator()
				menu_select.add_item(cmd_connected_perpendicular_faces)
				menu_select.add_item(cmd_connected_parallel_faces)
				menu_select.add_item(cmd_connected_planar_faces)
				menu_select.add_item(cmd_connected_faces_by_angle)
				menu_select.add_item(cmd_connected_by_material)
				menu_select.add_item(cmd_connected_by_back_material)
				menu_select.add_item(cmd_connected_by_layer)
				menu_select.add_item(cmd_connected_faces_by_area)
				menu_select.add_separator()
				menu_select.add_item(cmd_perpendicular_faces)
				menu_select.add_item(cmd_same_direction_faces)
				menu_select.add_item(cmd_parallel_faces)
				menu_select.add_item(cmd_opposite_faces)
				menu_select.add_item(cmd_faces_by_area)
			end
		end
		# Selection Filters
		if Sketchup.active_model.selection.length > 0
			menu_select_only = menu.add_submenu('Select Only')
			menu_select_only.add_item(cmd_filter_edges)
			menu_select_only.add_item(cmd_filter_faces)
			menu_select_only.add_item(cmd_filter_groups)
			menu_select_only.add_item(cmd_filter_components)
			menu_select_only.add_item(cmd_filter_guides)
			menu_select_only.add_item(cmd_filter_points)
			menu_select_only.add_item(cmd_filter_text)
			menu_select_only.add_item(cmd_filter_images)
			menu_select_only.add_item(cmd_filter_curves)
			menu_select_only.add_item(cmd_filter_arcs)
			menu_select_only.add_item(cmd_filter_circles)
			menu_select_only.add_item(cmd_filter_3dpoly)
			menu_select_only.add_item(cmd_filter_dim_lin)
			menu_select_only.add_item(cmd_filter_dim_rad)
			menu_select_only.add_separator()
			menu_select_only.add_item(cmd_filter_default_front_material)
			menu_select_only.add_item(cmd_filter_default_back_material)
			menu_select_only.add_separator()
			menu_select_only.add_item(cmd_filter_hidden)
			menu_select_only.add_item(cmd_filter_soft)
			menu_select_only.add_item(cmd_filter_smooth)
			#menu_select_only.add_separator()
			#menu_select_by_layer = menu_select_only.add_submenu('From Layer')
			#Select_Toys.filter_layer_menu(menu_select_by_layer)
			
			menu_deselect = menu.add_submenu('Deselect')
			menu_deselect.add_item(cmd_deselect_edges)
			menu_deselect.add_item(cmd_deselect_faces)
			menu_deselect.add_item(cmd_deselect_groups)
			menu_deselect.add_item(cmd_deselect_components)
			menu_deselect.add_item(cmd_deselect_guides)
			menu_deselect.add_item(cmd_deselect_points)
			menu_deselect.add_item(cmd_deselect_text)
			menu_deselect.add_item(cmd_deselect_images)
			menu_deselect.add_item(cmd_deselect_curves)
			menu_deselect.add_item(cmd_deselect_arcs)
			menu_deselect.add_item(cmd_deselect_circles)
			menu_deselect.add_item(cmd_deselect_3dpoly)
			menu_deselect.add_item(cmd_deselect_dim_lin)
			menu_deselect.add_item(cmd_deselect_dim_rad)
			menu_deselect.add_separator()
			menu_deselect.add_item(cmd_deselect_default_front_material)
			menu_deselect.add_item(cmd_deselect_default_back_material)
			menu_deselect.add_separator()
			menu_deselect.add_item(cmd_deselect_hidden)
			menu_deselect.add_item(cmd_deselect_soft)
			menu_deselect.add_item(cmd_deselect_smooth)
			#menu_deselect.add_separator()
			#menu_deselect_by_layer = menu_deselect.add_submenu('From Layer')
			#Select_Toys.deselect_layer_menu(menu_deselect_by_layer)
		end
	end
end

#-----------------------------------------------------------------------------
file_loaded('select_toys.rb')
#-----------------------------------------------------------------------------

module Select_Toys

	### COMPONENTS ### -------------------------------------------------------
	def self.select_active_instances
		model = Sketchup.active_model
		sel = model.selection
		
		proto = sel[0].definition
		
		##protos = []
		##sel.each { |e| protos << e.definition if e.respond_to :definition && protos.include?(e.definition)
		
		# Find all the components and add them to the selection
		model.active_entities.each { | ent | 
			if ent.kind_of?(Sketchup::ComponentInstance) && ent.definition.equal?(proto) && ent.layer.visible?
			##if ent.kind_of?(Sketchup::ComponentInstance) && protos.include?(ent.definition) && ent.layer.visible?
				sel.add(ent)
			end
		}
	end

	def self.select_instances
		# Selects all instances in the model across groups and components
		sel = Sketchup.active_model.selection
		sel.add( sel[0].definition.instances )
	end
	
	# Selects all instances from same layer
	def self.select_instances_from_same_layer
		model = Sketchup.active_model
		sel = model.selection
		
		layer = sel[0].layer
		proto = sel[0].definition
		
		proto.instances.each { | instance |
			if instance.layer == layer
				sel.add( instance )
			end
		}
	end
	
	# Selects active instances from same layer
	def self.select_active_instances_from_same_layer
		model = Sketchup.active_model
		sel = model.selection
		
		layer = sel[0].layer
		proto = sel[0].definition
		
		# Find all the components and add them to the selection
		model.active_entities.each { | ent | 
			if ent.kind_of?(Sketchup::ComponentInstance) && ent.definition.equal?(proto) && ent.layer == layer && ent.layer.visible?
				sel.add(ent)
			end
		}
	end

	
	### DYNAMIC COMPONENTS ### -----------------------------------------------
	def self.select_dc_active_instances
		model = Sketchup.active_model
		sel = model.selection
		
		dc_name = sel[0].definition.get_attribute('dynamic_attributes', '_name')
		
		# Find all the components and add them to the selection
		model.active_entities.each { | ent | 
			if ent.kind_of?(Sketchup::ComponentInstance) &&
			   ent.definition.get_attribute('dynamic_attributes', '_name') == dc_name &&
			   ent.layer.visible?
				sel.add(ent)
			end
		}
	end
	
	def self.select_dc_instances
		model = Sketchup.active_model
		sel = model.selection
		
		dc_name = sel[0].definition.get_attribute('dynamic_attributes', '_name')
		
		# Find all the components and add them to the selection
		model.definitions.each { | definition | 
			if definition.get_attribute('dynamic_attributes', '_name') == dc_name
				sel.add(definition.instances)
			end
		}
	end
	
	# Selects all instances from same layer
	def self.select_dc_instances_from_same_layer
		model = Sketchup.active_model
		sel = model.selection
		
		layer = sel[0].layer
		dc_name = sel[0].definition.get_attribute('dynamic_attributes', '_name')
		
		# Find all the components and add them to the selection
		model.definitions.each { | definition | 
			if definition.get_attribute('dynamic_attributes', '_name') == dc_name
				definition.instances.each  { |instance|
					sel.add(instance) if instance.layer == layer
				}
			end
		}
	end
	
	# Selects active instances from same layer
	def self.select_dc_active_instances_from_same_layer
		model = Sketchup.active_model
		sel = model.selection
		
		layer = sel[0].layer
		dc_name = sel[0].definition.get_attribute('dynamic_attributes', '_name')
		
		# Find all the components and add them to the selection
		model.active_entities.each { | ent | 
			if ent.kind_of?(Sketchup::ComponentInstance) &&
			   ent.definition.get_attribute('dynamic_attributes', '_name') == dc_name &&
			   ent.layer == layer &&
			   ent.layer.visible?
				sel.add(ent)
			end
		}
	end
	
	
	### GROUPS ### -----------------------------------------------------------
	
	def self.select_active_group_copies
		model = Sketchup.active_model
		sel = model.selection
		
		definition = sel[0].entities.parent
		
		# Find all the components and add them to the selection
		model.active_entities.each { | ent | 
			if ent.kind_of?(Sketchup::Group) && ent.entities.parent.equal?(definition) && ent.layer.visible?
				sel.add(ent)
			end
		}
	end
	
	def self.select_group_copies
		sel = Sketchup.active_model.selection
		sel.add(sel[0].entities.parent.instances)
	end
	
		# Selects all instances from same layer
	def self.select_group_copies_from_same_layer
		model = Sketchup.active_model
		sel = model.selection
		
		layer = sel[0].layer
		proto = sel[0].entities.parent
		
		proto.instances.each { | instance |
			if instance.layer == layer
				sel.add( instance )
			end
		}
	end
	
	# Selects active instances from same layer
	def self.select_active_group_copies_from_same_layer
		model = Sketchup.active_model
		sel = model.selection
		
		layer = sel[0].layer
		proto = sel[0].entities.parent
		
		# Find all the components and add them to the selection
		model.active_entities.each { | ent | 
			if ent.kind_of?(Sketchup::Group) && ent.entities.parent.equal?(proto) && ent.layer == layer
				sel.add(ent)
			end
		}
	end
	
	def self.convert_group_copies_to_components
		model = Sketchup.active_model
		sel = model.selection
		
		Helping_Hand.start_operation('Convert Copies to Components')
		
		# Find all the copies and make a prototype
		instances = sel[0].entities.parent.instances
		proto = instances[0].to_component
		
		# Iterate through the rest of the copies
		protodef = proto.definition
		instances.each { | inst |
			unless inst.deleted?
				# Get transformation and material
				trans = inst.transformation
				mat = inst.material
				layer = inst.layer
				# Add a new component
				new_comp = inst.parent.entities.add_instance(protodef, trans)
				new_comp.material = mat
				new_comp.layer = layer
				# Delete the old group
				inst.parent.entities.erase_entities inst
			end
		}
		
		model.commit_operation
		Sketchup.set_status_text(protodef.instances.length.to_s + ' group copies converted into components', SB_PROMPT)
	end
	
	
	### EDGES ### ------------------------------------------------------------
	
	# --- Select Loop from Edge ---
	def self.select_edge_loop
		puts "\n### Edge Loop ###"
		# Prepare first array to feed the loop
		edges = [Sketchup.active_model.selection[0]]
		used_edges = []
		while edges.length > 0
			edge = edges.pop
			used_edges << edge
			
			#puts "\n Edge: #{edge}"
			
			#edges << edge.start.edges[2] if edge.start.edges.length == 4 && !used_edges.include?(edge.start.edges[2])
			#edges << edge.end.edges[2] if edge.end.edges.length == 4 && !used_edges.include?(edge.end.edges[2])
			
			
			edge.vertices.each { |v|
				next unless v.edges.length == 4
				
				# Get hash with edges and their angles
				#puts "\n"
				v1 = (edge.start == v) ? edge.line[1].reverse : edge.line[1]
				angles = {}
				v.edges.each { |e|
					#puts e
					next if e == edge || used_edges.include?(e)
					# (!) Get vectors facing the vertex.
					v2 = (e.start == v) ? e.line[1].reverse : e.line[1]
					angles[e] = self.vector_full_angle_between(v1, v2, v)
				}
				#puts angles.inspect
				puts angles.sort{|a,b| a[1]<=>b[1]}.inspect
				angles = angles.sort{|a,b| a[1]<=>b[1]}
				#edges << angles[1][0] if angles.length == 3
			}
		end
		Sketchup.active_model.selection.clear
		Sketchup.active_model.selection.add(used_edges)
		puts "> #{used_edges.length} selected"
	end
	
	def self.vertex_edges(edge, vertex)
		return vertex.edges.reject { |e| e == edge }
	end
	
	# Return the full orientation (in 360 degrees instead of 180) of the two lines.
	# Going counter-clockwise.
	def self.vector_full_angle_between(vector1, vector2, p=nil)
		#puts "Vector #{vector1} - #{vector2}"
		# We need to find if 'vector' goes to the right or left of 'self'.
		# - We get the Z axis of the cross vector - ensuring it points upwards.
		return 0.0 if vector1 == vector2
		#return 0.0 if vector1 == vector2.reverse
		return 180.degrees if vector1 == vector2.reverse
		cross_vector = (vector1 * vector2).axes.z
		cross_vector.reverse! if cross_vector.z < 0
		unless p.nil?
			#to = 
			Sketchup.active_model.entities.add_cline(p, cross_vector)
		end
		# - Now we get the direction.
		# - direction > 0.0 = left
		# - direction < 0.0 = right
		direction = (vector1 * vector2) % cross_vector
		# Get the angle between the lines.
		angle = vector1.angle_between(vector2)
		# Correct the value returned to yield a 360 degree result.
		angle = 360.degrees - angle if direction < 0.0
		return angle
	end
	
	# --- Select Loop from Edge ---
	def self.select_loop_from_edge
		# Prepare first array to feed the loop
		edges = [Sketchup.active_model.selection[0]]
		while edges.length > 0
			edges = self.select_next_face(edges)
		end
	end
	
	def self.select_next_face(edges)
		sel = Sketchup.active_model.selection
		
		# Filter out the faces we've allready selected and faces
		# with more than four edges
		next_edges = [] # Collection of edges for next loop iteration
		# Iterate the edges we where given and find the next face for each of them
		# Once the face is found we get the oposite edge
		edges.each { | edge |
			# Filter out any face we allready have and any face with more than four sides
			faces = edge.faces.reject { | face |
				if sel.contains?(face) || face.edges.length > 4
					true
				else
					false
				end
			}
			
			# Add the new faces to the selection
			sel.add(faces)
			
			# Get the opposite edge and loop
			faces.each { | face |		
				other_edge = self.get_opposite_edge(edge, face)
				if other_edge != nil && !sel.contains?(other_edge)
					sel.add( other_edge ) 
					next_edges.push( other_edge )
				end
			}
		}
		return next_edges
	end
	
	def self.get_opposite_edge(edge, face)
		# Return the edge, in edges, which vertices isn't connected to edge.
		# This is effectivly the opposite edge in a face with four sides.
		edges = face.edges
		v = edge.vertices
		edges.each { |e|
			# Not sure if this is the best way to check this, but it simply checks
			# if any of the vertices from the tested edge exists in the array of
			# vertices from 'edge'
			if !v.include?(e.vertices[0]) && !v.include?(e.vertices[1])
				return e
			end
		}
		return nil
	end
	
	
	### FACES ### -----------------------------------------------------------
	# --- Select Connected Planar Faces ---
	def self.select_connected_planar_faces
		sel = Sketchup.active_model.selection
		face = sel[0]
		vector = face.normal
		
		edges = face.edges
		self.select_connected_faces(edges) { |face|
			!sel.contains?(face) && face.normal == vector
		}
	end
	
	# --- Select Connected Paralell Faces ---
	def self.select_connected_parallel_faces
		sel = Sketchup.active_model.selection
		face = sel[0]
		vector = face.normal
		
		face.all_connected.each { |e|
			sel.add(e) if e.kind_of?(Sketchup::Face) && e.normal.parallel?(vector)
		}
	end
	
	# --- Select Connected Faces by Area ---
	def self.select_connected_same_area
		sel = Sketchup.active_model.selection
		face = sel[0]
		
		face.all_connected.each { |e|
			sel.add(e) if e.kind_of?(Sketchup::Face) && e.area == face.area
		}
	end
	
	# --- Select Faces by Area ---
	def self.select_same_area
		sel = Sketchup.active_model.selection
		face = sel[0]
		
		Sketchup.active_model.active_entities.each { |e|
			sel.add(e) if e.kind_of?(Sketchup::Face) && e.area == face.area
		}
	end
	
	# --- Select Opposite Faces ---
	def self.select_opposite_faces
		sel = Sketchup.active_model.selection
		face = sel[0]
		vector = face.normal
		distance = nil
		op = face
		
		face.all_connected.each { |e|
			next if e == face
			if e.kind_of?(Sketchup::Face) && e.normal.reverse.samedirection?(vector)
				# Find distance to face, pick the closest (might be multiple)
				
				# We take the distance of the first vertex of the select face and measure the
				# distance to the current face we are testing.
				ds = face.vertices.first.position.distance_to_plane(e.plane)
				
				# If the face is closer than the one we currently have; we choose it.
				if distance == nil
					op = e
					distance = ds
				else
					if ds < distance
						op = e
						distance = ds 
					end
				end
			end
		}
		sel.add(op)
	end
	
	# --- Select Parallel Faces ---
	def self.select_parallel_faces
		sel = Sketchup.active_model.selection
		face = sel[0]
		vector = face.normal
		
		Sketchup.active_model.active_entities.each { |e|
			sel.add(e) if e.kind_of?(Sketchup::Face) && e.normal.parallel?(vector)
		}
	end
	
	# --- Select Faces in same direction ---
	def self.select_same_direction_faces
		sel = Sketchup.active_model.selection
		face = sel[0]
		vector = face.normal
		
		Sketchup.active_model.active_entities.each { |e|
			sel.add(e) if e.kind_of?(Sketchup::Face) && e.normal.samedirection?(vector)
		}
	end
	
	# --- Select Connected Perpendicular Faces ---
	def self.select_connected_perpendicular_faces
		sel = Sketchup.active_model.selection
		face = sel[0]
		vector = face.normal
		
		face.edges.each { |edge|
			edge.faces.each { |face|
				sel.add(face) if face.normal.perpendicular?(vector)
			}
		}
	end
	
	# --- Select Perpendicular Faces ---
	def self.select_perpendicular_faces
		sel = Sketchup.active_model.selection
		face = sel[0]
		vector = face.normal
		
		Sketchup.active_model.active_entities.each { |e|
			sel.add(e) if e.kind_of?(Sketchup::Face) && e.normal.perpendicular?(vector)
		}
	end

	
	### ENTITIES ###---------------------------------------------------------
	# --- Select Connected Entities By Material ---
	def self.select_connected_by_material(backside = false)
		sel = Sketchup.active_model.selection
		source = sel[0]
		
		if source.kind_of?(Sketchup::Face)
			source_material = (backside) ? source.back_material : source.material
		else
			source_material = source.material
		end
		
		edges = (source.kind_of?(Sketchup::Face)) ? source.edges : [source]
		self.select_connected_entities(edges) { |entity|
			if entity.kind_of?(Sketchup::Face)
				material = (backside) ? entity.back_material : entity.material
			else
				material = entity.material
			end
			!sel.contains?(entity) && material == source_material
		}
	end


	# --- Select Connected Entities By Layer ---
	def self.select_connected_by_layer
		sel = Sketchup.active_model.selection
		source = sel[0]
		layer = source.layer
		
		edges = (source.kind_of?(Sketchup::Face)) ? source.edges : [source]
		self.select_connected_entities(edges) { |entity|
			!sel.contains?(entity) && entity.layer == layer
		}
	end
	
	
	# --- Select Everything By Selected Layers ---
	def self.select_active_by_selected_layers
		sel = Sketchup.active_model.selection
		# Get list of layers to select.
		layers = []
		sel.each { |e| layers << e.layer unless layers.include?(e.layer) }
		# Select the entities.
		sel.clear
		Sketchup.active_model.active_entities.each { |e| sel.add(e) if layers.include?(e.layer) }
	end
	
	# --- Select Everything By Selected Layers ---
	def self.select_by_selected_layers
		sel = Sketchup.active_model.selection
		# Get list of layers to select.
		layers = []
		sel.each { |e| layers << e.layer unless layers.include?(e.layer) }
		# Select the entities.
		sel.clear
		self.all_entities { |e| sel.add(e) if layers.include?(e.layer) }
	end
	
	def self.all_entities
		Sketchup.active_model.entities.each { |entity| yield(entity) }
		Sketchup.active_model.definitions.each { |definition|
			next if definition.image? || definition.count_instances == 0
			definition.entities.each { |entity|
				yield(entity)
			}
		}
	end
	
	
	# --- Filter Layer
	def self.filter_layer_menu(menu)
		Sketchup.active_model.layers.each { |layer|
			menu.add_item( UI::Command.new(layer.name) {
				Sketchup.active_model.selection.clear
				Select_Toys.all_entities { |entity|
					Sketchup.active_model.selection.add(entity) if entity.layer == layer
				}
			} )
		}
	end
	
	def self.deselect_layer_menu(menu)
		Sketchup.active_model.layers.each { |layer|
			menu.add_item( UI::Command.new(layer.name) {
				Select_Toys.all_entities { |entity|
					Sketchup.active_model.selection.remove(entity) if entity.layer == layer
				}
			} )
		}
	end
	
	
	### SELECTION ### -------------------------------------------------------
	def self.deselect2()
		sel = Sketchup.active_model.selection
		sel.each { |ent|
			if yield(ent)
				sel.remove ent
			end
		}
	end
	
	def self.deselect(type)
		sel = Sketchup.active_model.selection
		sel.each { |ent|
			if ent.kind_of?(type)
				sel.remove ent
			end
		}
	end
	
	def self.deselect_typename(type)
		sel = Sketchup.active_model.selection
		sel.each { |ent|
			if ent.typename == type
				sel.remove ent
			end
		}
	end
	
	def self.deselect_drawingelement(typename)
		sel = Sketchup.active_model.selection
		sel.each { |ent|
			if ent.class == Sketchup::Drawingelement && ent.typename == typename
				sel.remove ent
			end
		}
	end
	
	def self.deselect_arc
		sel = Sketchup.active_model.selection
		sel.each { |ent|
			if ent.kind_of?(Sketchup::Edge) && ent.curve && ent.curve.kind_of?(Sketchup::ArcCurve)
				sel.remove ent
			end
		}
	end
	
	def self.deselect_curves
		sel = Sketchup.active_model.selection
		sel.each { |ent|
			if ent.kind_of?(Sketchup::Edge) && ent.curve
				sel.remove ent
			end
		}
	end
	
	def self.deselect_circle
		sel = Sketchup.active_model.selection
		sel.each { |ent|
			if ent.kind_of?(Sketchup::Edge) && ent.curve && ent.curve.kind_of?(Sketchup::ArcCurve) && self.is_circle?(ent.curve)
				sel.remove ent
			end
		}
	end
	
	def self.deselect_default_material(backface)
		sel = Sketchup.active_model.selection
		sel.each { |ent|
			if backface && ent.kind_of?(Sketchup::Face)
				sel.remove ent if ent.back_material == nil
			else
				sel.remove ent if ent.material == nil
			end
		}
	end
	
	# New improved
	def self.filter_selection2()
		sel = Sketchup.active_model.selection
		sel.each { |ent|
			if !yield(ent)
				sel.remove ent
			end
		}
	end
	
	def self.filter_selection(filter)
		sel = Sketchup.active_model.selection
		sel.each { |ent|
			if !ent.kind_of?(filter)
				sel.remove ent
			end
		}
	end
	
	def self.filter_selection_drawingelement(typename)
		sel = Sketchup.active_model.selection
		sel.each { |ent|
			unless ent.class == Sketchup::Drawingelement && ent.typename == typename
				sel.remove ent
			end
		}
	end
	
	def self.filter_selection_arc
		sel = Sketchup.active_model.selection
		sel.each { |ent|
			unless ent.kind_of?(Sketchup::Edge) && ent.curve && ent.curve.kind_of?(Sketchup::ArcCurve)
				sel.remove ent
			end
		}
	end
	
	def self.filter_selection_curves
		sel = Sketchup.active_model.selection
		sel.each { |ent|
			unless ent.kind_of?(Sketchup::Edge) && ent.curve
				sel.remove ent
			end
		}
	end
	
	def self.filter_selection_circle
		sel = Sketchup.active_model.selection
		sel.each { |ent|
			unless ent.kind_of?(Sketchup::Edge) && ent.curve && ent.curve.kind_of?(Sketchup::ArcCurve) && self.is_circle?(ent.curve)
				sel.remove ent
			end
		}
	end
	
	def self.filter_selection_default_material(backface=false)
		sel = Sketchup.active_model.selection
		sel.each { |ent|
			if backface && ent.kind_of?(Sketchup::Face)
				sel.remove ent if ent.back_material != nil
			else
				sel.remove ent if ent.material != nil
			end
		}
	end
	
	
	### SELECT CONNECTED ### -------------------------------------------------
	# Connected Faces
	def self.select_connected_faces(edges)
		sel = Sketchup.active_model.selection
		while faces = select_connected_faces_from_edges(edges)	
			# We now have a list of connected faces...
			edges.clear()
			faces.each { |face|
				# ...check each face if it's co-planar
				if yield(face)
					sel.add( face )
					edges += face.edges
				end
			}
		end
	end
	
	def self.select_connected_faces_from_edges(edges)
		faces = []
		edges.each { |edge|
			faces += edge.faces
		}
		return (faces.length > 0) ? faces : nil
	end
	
	# Connected Entities
	def self.select_connected_entities(edges)
		sel = Sketchup.active_model.selection
		while faces = select_connected_faces_from_edges(edges)	
			# We now have a list of connected faces...
			edges.clear()
			faces.each { |face|
				# ...check each face if it's co-planar
				if yield(face)
					sel.add( face )
					edges += face.edges
					# Add edges
					edges.each { |e|
						if yield(e)
							sel.add( e )
						end
					}
				end
			}
		end
	end
	
	
	### VALIDATIONS ### ------------------------------------------------------

	def self.selected_component?
		sel = Sketchup.active_model.selection
		
		# Ensure a component is selected and that there are
		# more instances
		if sel.length == 1 &&
		sel[0].kind_of?(Sketchup::ComponentInstance) &&
		sel[0].definition.instances.length > 1
			return true
		else
			return false
		end
	end
	
	def self.selected_dc_component?
		sel = Sketchup.active_model.selection
		
		# Ensure a component is selected and that there are
		# more instances
		if sel.length == 1 && sel[0].kind_of?(Sketchup::ComponentInstance)
			return true if sel[0].attribute_dictionary('dynamic_attributes') != nil
		else
			return false
		end
	end
	
	def self.selected_group_copies?
		sel = Sketchup.active_model.selection
		
		# Ensure a group with copies is selected and that it
		# has copies
		# (!) SU messes up the entity DB some times. We need to check if for this.
		#     group.entities.parent.include?(group) <- if not; track down real definition and
		#     check that for number of instances.
		if sel.length == 1 &&
		sel[0].kind_of?(Sketchup::Group) &&
		Helping_Hand.get_real_group_parent(sel[0]).instances.length > 1
		#sel[0].entities.parent.instances.length > 1
			return true
		else
			return false
		end
	end
	
	def self.selected_edge?
		sel = Sketchup.active_model.selection
		
		# Ensure an edge is selected
		if sel.length == 1 && sel[0].kind_of?(Sketchup::Edge)
			return true
		else
			return false
		end
	end
	
	def self.selected_edges?
		sel = Sketchup.active_model.selection
		
		# Ensure  edges are selected
		sel.each { |e|
			if !e.kind_of?(Sketchup::Edge)
				return false
			end
		}
		return true
	end
	
	def self.selected_face?
		sel = Sketchup.active_model.selection
		
		# Ensure an face is selected
		if sel.length == 1 && sel[0].kind_of?(Sketchup::Face)
			return true
		else
			return false
		end
	end
	
	def self.selected_faces?
		sel = Sketchup.active_model.selection
		
		# Ensure only faces are selected
		sel.each { |e|
			if !e.kind_of?(Sketchup::Face)
				return false
			end
		}
		return true
	end
	
	def self.is_circle?(arc_curve)
		# (i) A bug in SU makes extruded circles into Arcs with .end_angle of 720 degrees when reopening the file.
		# Instead of checking for 360 degrees exactly, we check for anything larger as well. A 2D arc
		# can't have more than 360 degrees.
		
		# Old check.
		#return (arc_curve.end_angle == Math::PI * 2) ? true : false
		
		# This doesn't work. Maybe due to rounding errors?
		#return (arc_curve.end_angle - arc_curve.start_angle >= 360.degrees) ? true : false
		
		return ((arc_curve.end_angle - arc_curve.start_angle).radians >= 360) ? true : false
	end
	
	### TOOLS ### ------------------------------------------------------------
	def self.tool(tool)
		Sketchup.active_model.select_tool(tool)
	end
	
	class Select_Connected_Faces_By_Angle_Tool
		@angle = 0.0
		@face = nil
		
		def activate
			Sketchup::set_status_text('Angle', SB_VCB_LABEL)
			Sketchup::set_status_text(0.0, SB_VCB_VALUE)
			
			@face = Sketchup.active_model.selection[0]
			@angle = 0.0
			self.select_by_angle()
		end
		
		def deactivate(view)
			view.invalidate
		end
		
		def draw(view)
			view.line_width = 5
			view.drawing_color = [255, 0, 0]
			@face.edges.each { |edge|
				view.draw_line(edge.start.position, edge.end.position)
			}
		end
		
		def draw_xxx(view)
			model = Sketchup.active_model
			sel = model.selection
			face = sel[0]
			
			ents = face.all_connected
			vector =  face.normal
			view.line_width = 5
			
			ents.each { |ent|
				if ent.kind_of?(Sketchup::Face)
					angle = vector.angle_between(ent.normal)
					ratio = angle / (Math::PI/2)
					
					red   = Integer(255 * ratio)
					green = Integer(255 - red)
					blue  = Integer(0)
					
					view.drawing_color = [red, green, blue]
					
					ent.edges.each { |edge|
						view.draw_line(edge.start.position, edge.end.position)
					}
				end
			}
		end
		
		def onUserText(text, view)
			@angle = text.to_f
			self.select_by_angle()
			Sketchup::set_status_text(text, SB_VCB_VALUE)
		end
		
		def select_by_angle
			sel = Sketchup.active_model.selection
			sel.clear()
			vector = @face.normal
			
			edges = @face.edges
			Select_Toys::select_connected_faces(edges) { |face|
				angle = vector.angle_between(face.normal).radians
				!sel.contains?(face) && (angle <= @angle)
			}
		end
	end
	
	
	### HELPER METHODS ### ---------------------------------------------------
	
	module Helping_Hand
		def self.start_operation(name)
			model = Sketchup.active_model
			# Make use of the SU7 speed boost with start_operation while
			# making sure it works in SU6.
			if Sketchup.version.split('.')[0].to_i >= 7
				model.start_operation(name, true)
			else
				model.start_operation(name)
			end
		end
		
		def self.get_real_group_parent(group)
			if group.entities.parent.instances.include?(group)
				return group.entities.parent
			else
				Sketchup.active_model.definitions.each { |definition|
					return definition if definition.instances.include?(group)
				}
			end
			return nil # error
		end
	end # module Helping_Hand
end # module Select_Toys